import { getDeserialize, GlobalRegistry, IterResult, Registry } from 'serde'
import { Deserialize, Forwarder, IDeserializer, IterableAccess, IVisitor, MapAccess, Visitor } from 'serde/de'

interface Reviver {
  <T, U>(key: string, value: T): U
}

type Entry = [any, any]

const unwrap = (de: IDeserializer) => de.deserializeAny(new Visitor())
const deserializer = (registry: Registry = GlobalRegistry) => <T, U>(_key: string, value: T) => {
  const de = getDeserialize(value, unwrap, registry)
  return de(new Forwarder(value)) as U
}

const parser = (reviver: Reviver) => <T, U>(value: T): U => {
  return JSON.parse(value as any, reviver)
}

export class JSONMapAccess extends MapAccess {
  private readonly _entries: Entry[]
  private index: number = -1

  constructor(entries: Entry[]) {
    super()

    this._entries = entries
  }

  static fromObject(value: object) {
    return new this(Object.entries(value))
  }

  nextKey<T>(seed?: Deserialize<T>): IteratorResult<T> {
    this.index += 1

    if (this.index >= this._entries.length) {
      return IterResult.Done()
    } else {
      const key = this._entries[this.index][0]
      const deser = seed != null ? seed(key) : key
      return IterResult.Next(deser)
    }
  }

  nextValue<T>(seed?: Deserialize<T>): IteratorResult<T> {
    if (this.index >= this._entries.length) {
      return IterResult.Done()
    } else {
      const value = this._entries[this.index][1]
      const deser = seed != null ? seed(value) : value
      return IterResult.Next(deser)
    }
  }
}

export class JSONIterableAccess extends IterableAccess {
  private readonly iterator: Iterator<any>

  constructor(iterator: Iterator<any>) {
    super()
    this.iterator = iterator
  }

  static fromIterable(iterable: Iterable<any>) {
    return new this(iterable[Symbol.iterator]())
  }

  nextElement<T, D extends Deserialize<T>>(seed?: D): IteratorResult<T> {
    const result = this.iterator.next()
    if (result.done) {
      return IterResult.Done()
    } else {
      const value = seed != null ? seed(result.value) : result.value
      return IterResult.Next(value)
    }
  }
}

export class JSONDeserializer implements IDeserializer {
  private readonly input: string
  private readonly parser: <T, U>(value: T) => U

  constructor(input: string, registry: Registry = GlobalRegistry) {
    this.input = input
    this.parser = parser(deserializer(registry))
  }

  static fromString(value: string, registry?: Registry) {
    return new this(value, registry)
  }

  deserializeAny<T>(_visitor: IVisitor<T>): T {
    return this.parser(this.input)
  }

  deserializeBoolean<T>(visitor: IVisitor<T>): T {
    return visitor.visitBoolean(this.parser(this.input))
  }

  deserializeNumber<T>(visitor: IVisitor<T>): T {
    return visitor.visitNumber(this.parser(this.input))
  }

  deserializeBigInt<T>(visitor: IVisitor<T>): T {
    return visitor.visitBigInt(this.parser(this.input))
  }

  deserializeString<T>(visitor: IVisitor<T>): T {
    return visitor.visitString(this.parser(this.input))
  }

  deserializeSymbol<T>(visitor: IVisitor<T>): T {
    return visitor.visitSymbol(this.parser(this.input))
  }

  deserializeNull<T>(visitor: IVisitor<T>): T {
    return visitor.visitNull()
  }

  deserializeObject<T>(visitor: IVisitor<T>): T {
    const value = this.parser(this.input) as object
    return visitor.visitObject(JSONMapAccess.fromObject(value))
  }

  deserializeIterable<T>(visitor: IVisitor<T>): T {
    const value = this.parser(this.input) as Iterable<any>
    return visitor.visitIterable(JSONIterableAccess.fromIterable(value))
  }

  deserializeFunction<T>(_visitor: IVisitor<T>): T {
    throw new Error('Method not implemented')
  }
}

