import { getSerialize, GlobalRegistry, Registry } from 'serde'
import { IdentitySerializer, ISerializer, serialize, SerializeIterable, SerializeObject, Serializer } from 'serde/ser'

interface Stringify {
  (value: any): string
}

type Replacer = (key: PropertyKey, value: any) => any

class JSONSerializeObject extends SerializeObject<string> {
  private readonly stringify: Stringify
  private value: Record<PropertyKey, any> = {}
  private currentKey?: string

  constructor(stringify: Stringify) {
    super()
    this.stringify = stringify
  }

  serializeKey(key: string): void {
    this.currentKey = key
  }

  serializeValue<U>(value: U): void {
    this.value[this.currentKey!] = value
    this.currentKey = undefined
  }

  end(): string {
    return this.stringify(this.value)
  }
}

class JSONSerializeIterable extends SerializeIterable<string> {
  private readonly stringify: Stringify
  private elements: any[] = []

  constructor(stringify: Stringify) {
    super()
    this.stringify = stringify
  }

  serializeElement<U>(value: U): void {
    this.elements.push(value)
  }

  end(): string {
    return this.stringify(this.elements)
  }
}

const id = <T, U>(_ser: ISerializer<T>, value: U) => value
const serializer = (registry: Registry) => <T>(_key: PropertyKey, value: T) => {
  const ser = getSerialize(value, id, registry)
  return ser(new IdentitySerializer(), value)
}

const stringify = (replacer: Replacer) => (value: any) => JSON.stringify(value, replacer)

export class JSONSerializer extends Serializer<string> {
  private stringify: (value: any) => string

  constructor(registry: Registry = GlobalRegistry) {
    super()
    this.stringify = stringify(serializer(registry))
  }

  serializeAny(value: any) { return this.stringify(value) }
  serializeBoolean(value: any) { return JSON.stringify(value) }
  serializeNumber(value: any) { return JSON.stringify(value) }
  serializeBigInt(value: any) { return JSON.stringify(value) }
  serializeString(value: any) { return JSON.stringify(value) }
  serializeSymbol(value: any) { return JSON.stringify(value) }
  serializeNull() { return JSON.stringify(null) }
  serializeIterable(_len: number) { return new JSONSerializeIterable(this.stringify) }
  serializeObject(_len: number) { return new JSONSerializeObject(this.stringify) }
  serializeClass(_name: string, _len: number) { return new JSONSerializeObject(this.stringify) }
}

